/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file stdout.c functions that deal with writing to stdout */

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "error.h"
#include "stdout.h"

/**
 * the pre_select function of the stdout task
 *
 * \param s the scheduler this task was registered to
 * \param t the task structure of the stdout task
 *
 * This function is always successful. If there is data available in the input
 * buffer, it adds \p STDOUT_FILENO to the write fd set of \a s.
 */
void stdout_pre_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = t->private_data;

	t->ret = 1;
	sot->check_fd = 0;
	if (!*sot->loaded) {
		if (*sot->input_eof) {
			t->ret = -E_STDOUT_EOF;
			s->timeout.tv_sec = 0;
			s->timeout.tv_usec = 1;
		}
		return;
	}
	sot->check_fd = 1;
	para_fd_set(STDOUT_FILENO, &s->wfds, &s->max_fileno);
}

/**
 * the post select function of the stdout task
 *
 * \param s the scheduler this task was registered to
 * \param t the task structure of the stdout task
 *
 * This function checks if \p STDOUT_FILENO was included by in the write fd set
 * of \a s during the previous pre_select call.  If yes, and \p STDOUT_FILENO
 * appeears to be writable, the data loaded in the input buffer is written to
 * stdout.
 */
void stdout_post_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = t->private_data;
	ssize_t ret;

	t->ret = 1;
	if (!sot->check_fd) {
		if (*sot->input_eof)
			t->ret = -E_STDOUT_EOF;
		return;
	}
	if (!FD_ISSET(STDOUT_FILENO, &s->wfds))
		return;
	t->ret = -E_STDOUT_WRITE;
	ret = write(STDOUT_FILENO, sot->buf, *sot->loaded);
	if (ret <= 0)
		return;
	*sot->loaded -= ret;
	if (*sot->loaded)
		memmove(sot->buf, sot->buf + ret, *sot->loaded);
	t->ret = 1;
}

static void stdout_default_event_handler(struct task *t)
{
	PARA_NOTICE_LOG("%p: %s\n", t, PARA_STRERROR(-t->ret));
	unregister_task(t);
}

/**
 * initialize a stdout task structure with default values
 *
 * \param sot the stdout task structure
 *
 * This fills in the pre/post select function poinzters of the task structure
 * given by \a sot. It also sets up a default error handler which unregisters
 * the task on errors and clears the eof flag of \a sot.
 */
void stdout_set_defaults(struct stdout_task *sot)
{
	sot->task.private_data = sot;
	sot->task.pre_select = stdout_pre_select;
	sot->task.post_select = stdout_post_select;
	sot->task.event_handler = stdout_default_event_handler;
	sot->eof = 0;
	mark_fd_nonblock(STDOUT_FILENO);
	sprintf(sot->task.status, "stdout writer");
}
